@@ -1,7 +1,7 @@
 /*
  * Global program option structure and the parsing function prototype.
  *
- * Copyright 2002-2008, 2010, 2012-2015, 2017, 2021, 2023-2024 Andrew Wood
+ * Copyright 2002-2008, 2010, 2012-2015, 2017, 2021, 2023-2025 Andrew Wood
  *
  * License GPLv3+: GNU GPL version 3 or later; see `docs/COPYING'.
  */
@@ -23,6 +23,16 @@ extern "C" {
 struct opts_s;
 typedef struct opts_s *opts_t;
 
+/*
+ * Overall program actions to select.
+ */
+typedef enum {
+	PV_ACTION_NOTHING,		/* do nothing, and exit */
+	PV_ACTION_TRANSFER,		/* transfer data */
+	PV_ACTION_STORE_AND_FORWARD,	/* store to file, then output from it */
+	PV_ACTION_WATCHFD		/* watch process file descriptors */
+} pvaction_t;
+
 /*
  * Structure describing run-time options.
  *
@@ -39,6 +49,8 @@ struct opts_s {
 	/*@keep@*/ /*@null@*/ char *pidfile; /* PID file, if any */
 	/*@keep@*/ /*@null@*/ char *store_and_forward_file; /* store and forward file, if any */
 	/*@keep@*/ /*@null@*/ char *extra_display; /* extra display specifier, if any */
+	/*@keep@*/ /*@null@*/ pid_t *watchfd_pid;  /* array of processes to watch fds of */
+	/*@keep@*/ /*@null@*/ int *watchfd_fd;  /* array of fds to watch in each one (0=all) */
 	/*@keep@*/ /*@null@*/ const char **argv;   /* array of non-option arguments */
 	size_t lastwritten;            /* show N bytes last written */
 	off_t rate_limit;              /* rate limit, in bytes per second */
@@ -47,14 +59,14 @@ struct opts_s {
 	off_t error_skip_block;        /* skip block size, 0 for adaptive */
 	pid_t remote;                  /* PID of pv to update settings of */
 	unsigned int skip_errors;      /* skip read errors counter */
-	pid_t watch_pid;	       /* process to watch fds of */
-	int watch_fd;		       /* fd to watch */
 	unsigned int average_rate_window; /* time window in seconds for average rate calculations */
 	unsigned int width;            /* screen width */
 	unsigned int height;           /* screen height */
 	unsigned int argc;             /* number of non-option arguments */
 	unsigned int argv_length;      /* allocated array size */
-	bool do_nothing;               /* exit-without-doing-anything flag */
+	unsigned int watchfd_count;	       /* number of watchfd items */
+	unsigned int watchfd_length;	       /* allocated array size */
+	pvaction_t action;	       /* the program action to perform */
 	bool progress;                 /* progress bar flag */
 	bool timer;                    /* timer flag */
 	bool eta;                      /* ETA flag */
@@ -77,6 +89,7 @@ struct opts_s {
 	bool stop_at_size;             /* set if we stop at "size" bytes */
 	bool sync_after_write;         /* set if we sync after every write */
 	bool direct_io;                /* set if O_DIRECT is to be used */
+	bool sparse_output;            /* set if we leave holes in the output */
 	bool discard_input;            /* set to write nothing to output */
 	bool show_stats;	       /* set to write statistics at the end */
 	bool width_set_manually;       /* width was set manually, not detected */
