@@ -60,7 +60,7 @@ extern "C" {
 /*
  * Structure for data shared between multiple "pv -c" instances.
  */
-struct pvcursorstate_s {
+struct pvipccursorstate_s {
 	int y_topmost;		/* terminal row of topmost "pv" instance */
 	bool tty_tostop_added;	/* whether any instance had to set TOSTOP on the terminal */
 };
@@ -131,6 +131,8 @@ typedef uint16_t pvdisplay_bytecount_t;
 typedef uint16_t pvdisplay_width_t;
 #define PVDISPLAY_WIDTH_MAX (65535)	/* UINT16_MAX */
 
+/* String pointer, that is the only pointer to this resource, that can be null. */
+typedef /*@only@*/ /*@null@*/ char * nullable_string_t;
 
 /*
  * Structure for holding PV internal state. Opaque outside the PV library.
@@ -141,25 +143,37 @@ struct pvstate_s {
 	/******************
 	 * Program status *
 	 ******************/
-	struct {
-		/*@only@*/ char *program_name;	 /* program name for error reporting */
+	struct pvprogramstatus_s {
 		char cwd[PV_SIZEOF_CWD];	 /* current working directory for relative path */
 		int current_input_file;		 /* index of current file being read */
 		int exit_status; 		 /* exit status to give (0=OK) */
+		bool terminal_supports_utf8;	 /* whether the terminal supports UTF-8 */
+		bool terminal_supports_colour;	 /* whether the terminal supports colour */
+		bool checked_colour_support;	 /* whether we have checked colour support yet */
 	} status;
 
 	/***************
 	 * Input files *
 	 ***************/
-	struct {
-		/*@only@*/ /*@null@*/ char **filename; /* input filenames */
+	struct pvinputfiles_s {
+		/*@only@*/ /*@null@*/ nullable_string_t *filename; /* input filenames */
 		unsigned int file_count;	 /* number of input files */
 	} files;
 
+	/*********************************
+	 * Items to watch with --watchfd *
+	 *********************************/
+	struct pvwatchspec_s {
+		/*@only@*/ /*@null@*/ pid_t *pid; /* array of processes to watch fds of */
+		/*@only@*/ /*@null@*/ int *fd;	/* array of fds to watch in each one (0=all) */
+		unsigned int count;	/* number of items in these arrays */
+		bool multiple_pids;	/* true if more than one distinct PID */
+	} watchfd;
+
 	/*******************
 	 * Program control *
 	 *******************/
-	struct {
+	struct pvcontrol_s {
 		char default_format[PV_SIZEOF_DEFAULT_FORMAT];	 /* default format string */
 		double interval;                 /* interval between updates */
 		double delay_start;              /* delay before first display */
@@ -172,9 +186,7 @@ struct pvstate_s {
 		off_t rate_limit;                /* rate limit, in bytes per second */
 		size_t target_buffer_size;       /* buffer size (0=default) */
 		off_t size;                      /* total size of data */
-		pid_t watch_pid;		 /* process to watch fds of */
 		unsigned int skip_errors;        /* skip read errors counter */
-		int watch_fd;			 /* fd to watch */
 		int output_fd;                   /* fd to write output to */
 		unsigned int average_rate_window; /* time window in seconds for average rate calculations */
 		unsigned int history_interval;	 /* seconds between each average rate calc history entry */
@@ -191,16 +203,14 @@ struct pvstate_s {
 		bool decimal_units;		 /* use decimal prefixes */
 		bool null_terminated_lines;      /* lines are null-terminated */
 		bool no_display;                 /* do nothing other than pipe data */
+		bool no_splice;                  /* never use splice() */
 		bool stop_at_size;               /* set if we stop at "size" bytes */
 		bool sync_after_write;           /* set if we sync after every write */
 		bool direct_io;                  /* set if O_DIRECT is to be used */
 		bool direct_io_changed;          /* set when direct_io is changed */
-		bool no_splice;                  /* never use splice() */
+		bool sparse_output;		 /* set if we leave holes in the output */
 		bool discard_input;              /* write nothing to stdout */
 		bool show_stats;		 /* show statistics on exit */
-		bool can_display_utf8;		 /* whether UTF-8 output is permitted */
-		bool can_display_colour;	 /* whether the terminal supports colour */
-		bool checked_colour_support;	 /* whether we have checked colour support yet */
 		bool width_set_manually;	 /* width was set manually, not detected */
 		bool height_set_manually;	 /* height was set manually, not detected */
 	} control;
@@ -208,7 +218,7 @@ struct pvstate_s {
 	/*******************
 	 * Signal handling *
 	 *******************/
-	struct {
+	struct pvsignal_s {
 		/* old signal handlers to restore in pv_sig_fini(). */
 		struct sigaction old_sigpipe;
 		struct sigaction old_sigttou;
@@ -233,7 +243,7 @@ struct pvstate_s {
 	/*******************
 	 * Transient flags *
 	 *******************/
-	struct {
+	struct pvtransientflags_s {
 		volatile sig_atomic_t reparse_display;	 /* whether to re-check format string */
 		volatile sig_atomic_t terminal_resized;	 /* whether we need to get term size again */
 		volatile sig_atomic_t trigger_exit;	 /* whether we need to abort right now */
@@ -241,7 +251,7 @@ struct pvstate_s {
 		volatile sig_atomic_t suspend_stderr;	 /* whether writing to stderr is suspended */
 		volatile sig_atomic_t skip_next_sigcont; /* whether to ignore the next SIGCONT */
 		volatile sig_atomic_t pipe_closed;	 /* whether the output pipe was closed */
-	} flag;
+	} flags;
 
 	/*****************
 	 * Display state *
@@ -295,7 +305,7 @@ struct pvstate_s {
 		bool colour_permitted;		 /* whether colour is permitted for this display */
 		bool sgr_code_active;		 /* set while SGR code is active in a display line */
 		bool final_update;		 /* set internally on the final update */
-		bool display_visible;		 /* set once anything written to terminal */
+		bool output_produced;		 /* set once anything written to terminal */
 
 	} display;
 
@@ -305,7 +315,7 @@ struct pvstate_s {
 	/************************************
 	 * Calculated state of the transfer *
 	 ************************************/
-	struct {
+	struct pvtransfercalc_s {
 		long double transfer_rate;	 /* calculated transfer rate */
 		long double average_rate;	 /* calculated average transfer rate */
 
@@ -337,10 +347,10 @@ struct pvstate_s {
 	/********************
 	 * Cursor/IPC state *
 	 ********************/
-	struct {
+	struct pvcursorstate_s {
 		char lock_file[PV_SIZEOF_CRS_LOCK_FILE];
 #ifdef HAVE_IPC
-		/*@keep@*/ /*@null@*/ struct pvcursorstate_s *shared; /* data shared between instances */
+		/*@keep@*/ /*@null@*/ struct pvipccursorstate_s *shared; /* data shared between instances */
 		int shmid;		 /* ID of our shared memory segment */
 		int pvcount;		 /* number of `pv' processes in total */
 		int pvmax;		 /* highest number of `pv's seen */
@@ -353,6 +363,7 @@ struct pvstate_s {
 #ifdef HAVE_IPC
 		bool noipc;		 /* set if we can't use IPC */
 #endif				/* HAVE_IPC */
+		bool disable;		 /* set if cursor positioning can't be used */
 	} cursor;
 
 	/*******************
@@ -373,7 +384,7 @@ struct pvstate_s {
 	 * is the offset in the buffer that we've written data up to.  It
 	 * will always be less than or equal to read_position.
 	 */
-	struct {
+	struct pvtransferstate_s {
 		long double elapsed_seconds;	 /* how long we have been transferring data for */
 		/*@only@*/ /*@null@*/ char *transfer_buffer;	 /* data transfer buffer */
 		size_t buffer_size;		 /* size of buffer */
@@ -385,6 +396,7 @@ struct pvstate_s {
 
 		size_t written_but_not_consumed; /* bytes in the output pipe, unread */
 
+		off_t total_bytes_read;		 /* total bytes read */
 		off_t total_written;		 /* total bytes or lines written */
 		off_t transferred;		 /* amount transferred (written - unconsumed) */
 
@@ -426,19 +438,47 @@ struct pvstate_s {
 		bool splice_used;
 #endif
 		bool read_error_warning_shown;
+		bool output_not_seekable;	/* set if lseek() fails on output */
 	} transfer;
 };
 
+typedef struct pvprogramstatus_s *pvprogramstatus_t;
+typedef struct pvinputfiles_s *pvinputfiles_t;
+typedef struct pvcontrol_s *pvcontrol_t;
+typedef struct pvsignal_s *pvsignal_t;
+typedef struct pvtransientflags_s *pvtransientflags_t;
 typedef struct pvdisplay_s *pvdisplay_t;
 typedef struct pvdisplay_segment_s *pvdisplay_segment_t;
+typedef struct pvtransfercalc_s *pvtransfercalc_t;
+typedef struct pvcursorstate_s *pvcursorstate_t;
+typedef struct pvtransferstate_s *pvtransferstate_t;
+
+/*
+ * Read-only counterparts to the above structure pointers, to be used in
+ * function declarations where the function definitely shouldn't be altering
+ * the contents of the structure.
+ */
+typedef const struct pvprogramstatus_s * readonly_pvprogramstatus_t;
+typedef const struct pvinputfiles_s * readonly_pvinputfiles_t;
+typedef const struct pvcontrol_s * readonly_pvcontrol_t;
+typedef const struct pvsignal_s * readonly_pvsignal_t;
+typedef const struct pvtransientflags_s * readonly_pvtransientflags_t;
+typedef const struct pvdisplay_s * readonly_pvdisplay_t;
+typedef const struct pvdisplay_segment_s * readonly_pvdisplay_segment_t;
+typedef const struct pvtransfercalc_s * readonly_pvtransfercalc_t;
+typedef const struct pvcursorstate_s * readonly_pvcursorstate_t;
+typedef const struct pvtransferstate_s * readonly_pvtransferstate_t;
 
 /*
  * Structure containing the parameters used by formatters.
  */
 struct pvformatter_args_s {
-	/*@dependent@*/ pvstate_t state;		/* overall state */
 	/*@dependent@*/ pvdisplay_t display;		/* the display being updated */
 	/*@dependent@*/ pvdisplay_segment_t segment;	/* the segment of the display */
+	/*@dependent@*/ readonly_pvprogramstatus_t status;	/* program status */
+	/*@dependent@*/ readonly_pvcontrol_t control;		/* control settings */
+	/*@dependent@*/ readonly_pvtransferstate_t transfer;	/* transfer state */
+	/*@dependent@*/ readonly_pvtransfercalc_t calc;		/* calculated transfer state */
 	/*@dependent@*/ char *buffer;			/* buffer to write formatted segments into */
 	pvdisplay_bytecount_t buffer_size;		/* size of the buffer */
 	pvdisplay_bytecount_t offset;			/* current write position in the buffer */
@@ -460,29 +500,39 @@ struct pvdisplay_component_s {
 };
 
 
+/*
+ * Structure defining the current state of a single watched file descriptor.
+ */
 struct pvwatchfd_s {
+	struct pvtransientflags_s flags;	/* transient flags */
+	struct pvtransferstate_s transfer;	/* transfer state */
+	struct pvtransfercalc_s calc;	 /* calculated transfer state */
+	struct pvdisplay_s display;	 /* display data */
 #ifdef __APPLE__
 #else
 	char file_fdinfo[PV_SIZEOF_FILE_FDINFO]; /* path to /proc fdinfo file */
 	char file_fd[PV_SIZEOF_FILE_FD];	 /* path to /proc fd symlink  */
 #endif
 	char file_fdpath[PV_SIZEOF_FILE_FDPATH]; /* path to file that was opened */
-	char display_name[PV_SIZEOF_DISPLAY_NAME]; /* name to show on progress bar */
+	/*@keep@ */ char display_name[PV_SIZEOF_DISPLAY_NAME]; /* name to show on progress bar */
 	struct stat sb_fd;		 /* stat of fd symlink */
 	struct stat sb_fd_link;		 /* lstat of fd symlink */
 	off_t size;			 /* size of whole file, 0 if unknown */
 	off_t position;			 /* position last seen at */
 	struct timespec start_time;	 /* time we started watching the fd */
-	/*@null@*/ pvstate_t state;	 /* state object for flags and display */
-	pid_t watch_pid;		 /* PID to watch */
-	int watch_fd;			 /* fd to watch, -1 = not displayed */
+	struct timespec end_time;	 /* time the fd was marked as closed */
+	pid_t watch_pid;		 /* PID the fd belongs to */
+	int watch_fd;			 /* fd to watch */
+	bool closed;			 /* true once the fd is closed */
+	bool displayable;		 /* false if not displayable */
+	bool unused;			 /* true if free for re-use */
 };
 typedef struct pvwatchfd_s *pvwatchfd_t;
 
-void pv_error(pvstate_t, char *, ...);
+void pv_error(char *, ...);
 
 int pv_main_loop(pvstate_t);
-void pv_calculate_transfer_rate(pvstate_t, bool);
+void pv_calculate_transfer_rate(pvtransfercalc_t, readonly_pvtransferstate_t, readonly_pvcontrol_t, readonly_pvdisplay_t, bool);
 
 long pv_bound_long(long, long, long);
 long pv_seconds_remaining(const off_t, const off_t, const long double);
@@ -534,20 +584,36 @@ pvdisplay_bytecount_t pv_formatter_previous_line(pvformatter_args_t);
 pvdisplay_bytecount_t pv_formatter_name(pvformatter_args_t);
 pvdisplay_bytecount_t pv_formatter_sgr(pvformatter_args_t);
 
-bool pv_format(pvstate_t, /*@null@*/ const char *, pvdisplay_t, bool, bool);
-void pv_display(pvstate_t, bool);
+bool pv_format (pvprogramstatus_t, readonly_pvcontrol_t,
+		readonly_pvtransferstate_t, readonly_pvtransfercalc_t,
+		/*@null@ */ const char *, pvdisplay_t, bool, bool);
+void pv_display (pvprogramstatus_t, readonly_pvcontrol_t, pvtransientflags_t,
+		 readonly_pvtransferstate_t, pvtransfercalc_t,
+		 pvcursorstate_t, pvdisplay_t, /*@null@ */ pvdisplay_t, bool);
+
 ssize_t pv_transfer(pvstate_t, int, bool *, bool *, off_t, long *);
 int pv_next_file(pvstate_t, unsigned int, int);
 /*@keep@*/ const char *pv_current_file_name(pvstate_t);
 
-void pv_write_retry(int, const char *, size_t);
-void pv_tty_write(pvstate_t, const char *, size_t);
+unsigned int pv_update_calc_average_rate_window(pvtransfercalc_t, unsigned int);
+void pv_reset_calc(pvtransfercalc_t);
+void pv_reset_transfer(pvtransferstate_t);
+void pv_reset_flags(pvtransientflags_t);
+void pv_reset_display(pvdisplay_t);
+void pv_reset_watchfd(pvwatchfd_t);
+void pv_freecontents_display(pvdisplay_t);
+void pv_freecontents_transfer(pvtransferstate_t);
+void pv_freecontents_calc(pvtransfercalc_t);
+void pv_freecontents_watchfd(pvwatchfd_t);
 
-void pv_crs_fini(pvstate_t);
-void pv_crs_init(pvstate_t);
-void pv_crs_update(pvstate_t, const char *);
+void pv_write_retry(int, const char *, size_t);
+void pv_tty_write(readonly_pvtransientflags_t, const char *, size_t);
+
+void pv_crs_fini(pvcursorstate_t, readonly_pvcontrol_t, pvtransientflags_t);
+void pv_crs_init(pvcursorstate_t, readonly_pvcontrol_t, pvtransientflags_t);
+void pv_crs_update(pvcursorstate_t, readonly_pvcontrol_t, pvtransientflags_t, const char *);
 #ifdef HAVE_IPC
-void pv_crs_needreinit(pvstate_t);
+void pv_crs_needreinit(pvcursorstate_t);
 #endif
 
 void pv_sig_allowpause(void);
@@ -555,14 +621,14 @@ void pv_sig_checkbg(void);
 void pv_sig_nopause(void);
 
 void pv_remote_init(pvstate_t);
-void pv_remote_check(pvstate_t);
+bool pv_remote_check(pvstate_t);
 void pv_remote_fini(pvstate_t);
 int pv_remote_set(pvstate_t);
 
 int pv_watchfd_info(pvstate_t, pvwatchfd_t, bool);
 bool pv_watchfd_changed(pvwatchfd_t);
 off_t pv_watchfd_position(pvwatchfd_t);
-int pv_watchpid_scanfds(pvstate_t, pid_t, int *, pvwatchfd_t *, int *);
+int pv_watchpid_scanfds(pvstate_t, pid_t, int, int *, pvwatchfd_t *);
 void pv_watchpid_setname(pvstate_t, pvwatchfd_t);
 
 #ifdef __cplusplus
