@@ -175,10 +175,15 @@ void pv_nanosleep(long long);
  * Main PV functions.
  */
 
+/*
+ * Set the prefix (program name) for any PV error messages.
+ */
+extern void pv_set_error_prefix(/*@unique@ */ const char *);
+
 /*
  * Create a new state structure, and return it, or 0 (NULL) on error.
  */
-extern /*@null@*/ /*@only@*/ pvstate_t pv_state_alloc(const char *);
+extern /*@null@*/ /*@only@*/ pvstate_t pv_state_alloc(void);
 
 /*
  * Clear the calculated parts of a state structure.
@@ -216,6 +221,7 @@ extern void pv_state_error_skip_block_set(pvstate_t, off_t);
 extern void pv_state_stop_at_size_set(pvstate_t, bool);
 extern void pv_state_sync_after_write_set(pvstate_t, bool);
 extern void pv_state_direct_io_set(pvstate_t, bool);
+extern void pv_state_sparse_output_set(pvstate_t, bool);
 extern void pv_state_rate_limit_set(pvstate_t, off_t);
 extern void pv_state_target_buffer_size_set(pvstate_t, size_t);
 extern void pv_state_no_splice_set(pvstate_t, bool);
@@ -228,13 +234,12 @@ extern void pv_state_name_set(pvstate_t, /*@null@*/ const char *);
 extern void pv_state_default_bar_style_set(pvstate_t, /*@null@ */ const char *);
 extern void pv_state_format_string_set(pvstate_t, /*@null@*/ const char *);
 extern void pv_state_extra_display_set(pvstate_t, /*@null@*/ const char *);
-extern void pv_state_watch_pid_set(pvstate_t, pid_t);
-extern void pv_state_watch_fd_set(pvstate_t, int);
 extern void pv_state_output_set(pvstate_t, int, const char *);
 extern void pv_state_average_rate_window_set(pvstate_t, unsigned int);
-extern void pv_state_set_can_display_utf8(pvstate_t, bool);
+extern void pv_state_set_terminal_supports_utf8(pvstate_t, bool);
 
 extern void pv_state_inputfiles(pvstate_t, unsigned int, const char **);
+extern void pv_state_watchfds(pvstate_t, unsigned int, const pid_t *, const int *);
 
 /*
  * Work out whether we are in the foreground.
@@ -269,15 +274,10 @@ extern bool pv_sigusr2_received(pvstate_t, pid_t *);
 extern int pv_main_loop(pvstate_t);
 
 /*
- * Watch the selected file descriptor of the selected process.
+ * Watch the selected file descriptors of the selected processes.
  */
 extern int pv_watchfd_loop(pvstate_t);
 
-/*
- * Watch the selected process.
- */
-extern int pv_watchpid_loop(pvstate_t);
-
 /*
  * Shut down signal handlers after running the main loop.
  */
