@@ -351,13 +351,16 @@ void display_help(void)
 		{ "-K", "--direct-io", NULL,
 		 N_("use direct I/O to bypass cache"),
 		 { 0, 0, 0, 0} },
+		{ "-O", "--sparse", NULL,
+		 N_("try to seek instead of writing null bytes"),
+		 { 0, 0, 0, 0} },
 		{ "-X", "--discard", NULL,
 		 N_("discard input instead of writing to output"),
 		 { 0, 0, 0, 0} },
 		{ "-U", "--store-and-forward", N_("FILE"),
 		 N_("write all input to FILE before writing to output"),
 		 { 0, 0, 0, 0} },
-		{ "-d", "--watchfd", N_("PID[:FD]"),
+		{ "-d", "--watchfd", N_("PID[:FD]|=NAME|@LISTFILE"),
 		 N_("watch file FD opened by process PID"),
 		 { 0, 0, 0, 0} },
 #ifdef PV_REMOTE_CONTROL
