@@ -27,6 +27,7 @@
 #endif
 #include <fcntl.h>
 #include <unistd.h>
+#include <sys/wait.h>
 #ifdef HAVE_GETOPT_H
 #include <getopt.h>
 #endif
@@ -34,6 +35,14 @@
 
 void display_help(void);
 void display_version(void);
+static bool opts_watchfd_parse(opts_t, const char *, /*@null@ */ const char *, unsigned int);
+
+
+/*
+ * splint note about mustfreefresh: the gettext calls made by _() cause
+ * memory leak warnings, but in these cases it's unavoidable, and mitigated
+ * by the fact we only translate each string once.
+ */
 
 
 /*
@@ -57,10 +66,16 @@ void opts_free( /*@only@ */ opts_t opts)
 		free(opts->pidfile);
 	if (NULL != opts->output)
 		free(opts->output);
+	if (NULL != opts->default_bar_style)
+		free(opts->default_bar_style);
 	if (NULL != opts->store_and_forward_file)
 		free(opts->store_and_forward_file);
 	if (NULL != opts->extra_display)
 		free(opts->extra_display);
+	if (NULL != opts->watchfd_pid)
+		free(opts->watchfd_pid);
+	if (NULL != opts->watchfd_fd)
+		free(opts->watchfd_fd);
 	if (NULL != opts->argv)
 		free(opts->argv);
 	/*@+keeptrans@ */
@@ -102,6 +117,300 @@ bool opts_add_file(opts_t opts, const char *filename)
 	return true;
 }
 
+/*
+ * Add a process ID and file descriptor to the list of items to watch with
+ * --watchfd, returning false on error.
+ */
+static bool opts_watchfd_add_item(opts_t opts, pid_t pid, int fd)
+{
+	/*@-branchstate@ */
+	if ((opts->watchfd_count >= opts->watchfd_length) || (NULL == opts->watchfd_pid)) {
+		opts->watchfd_length = opts->watchfd_count + 10;
+		/*@-keeptrans@ */
+		opts->watchfd_pid = realloc(opts->watchfd_pid, opts->watchfd_length * sizeof(pid_t));
+		opts->watchfd_fd = realloc(opts->watchfd_fd, opts->watchfd_length * sizeof(int));
+		/*@+keeptrans@ */
+		if ((NULL == opts->watchfd_pid) || (NULL == opts->watchfd_fd)) {
+			fprintf(stderr, "%s: %s\n", opts->program_name, strerror(errno));
+			opts->watchfd_length = 0;
+			opts->watchfd_count = 0;
+			return false;
+		}
+	}
+	/*@+branchstate@ */
+
+	/*
+	 * splint notes: we turned off "branchstate" and "keeptrans" above
+	 * because of the same reason as in opts_add_file().
+	 */
+
+	opts->watchfd_pid[opts->watchfd_count] = pid;
+	opts->watchfd_fd[opts->watchfd_count] = fd;
+	opts->watchfd_count++;
+
+	return true;
+}
+
+
+/*
+ * Look for processes with the given name, and add all of them to the list
+ * of watched PIDs.  Returns false on error.
+ */
+static bool opts_watchfd_processname(opts_t opts, const char *process_name)
+{
+	int fds[2];
+	pid_t pid;
+	FILE *fptr;
+	char buffer[1024];		 /* flawfinder: ignore */
+	pid_t waited_pid;
+	int pid_status;
+	bool ok;
+
+	/*
+	 * flawfinder: buffer is zeroed before each use, and fgets() is
+	 * passed one less than its size so the string functions in the loop
+	 * always find a null byte at the end.
+	 */
+
+	/* Pipe for communicating with pgrep. */
+	if (pipe(fds) < 0) {
+		fprintf(stderr, "%s: %s\n", opts->program_name, strerror(errno));
+		return false;
+	}
+
+	/* Subprocess in which to run pgrep. */
+	pid = (pid_t) fork();
+	if (pid < 0) {
+		fprintf(stderr, "%s: %s\n", opts->program_name, strerror(errno));
+		(void) close(fds[0]);
+		(void) close(fds[1]);
+		return false;
+	} else if (0 == pid) {
+		int nullfd;
+
+		/* Child process - close stdin, set up stdout, run pgrep. */
+
+		/* Replace stdin with /dev/null. */
+		nullfd = open("/dev/null", O_RDONLY);	/* flawfinder: ignore */
+		/* flawfinder: /dev/null is trusted. */
+		if (nullfd < 0) {
+			fprintf(stderr, "%s: %s: %s\n", opts->program_name, "/dev/null", strerror(errno));
+			exit(EXIT_FAILURE);
+		}
+		if (dup2(nullfd, STDIN_FILENO) < 0) {
+			perror("dup2");
+			exit(EXIT_FAILURE);
+		}
+		(void) close(nullfd);
+
+		/* Replace stdout with the write end of the pipe. */
+		if (dup2(fds[1], STDOUT_FILENO) < 0) {
+			perror("dup2");
+			exit(EXIT_FAILURE);
+		}
+		(void) close(fds[1]);
+
+		/* Close the read end of the pipe. */
+		(void) close(fds[0]);
+
+		/* Run pgrep. */
+		if (execlp("pgrep", "pgrep", process_name, NULL) < 0) {	/* flawfinder: ignore */
+			perror("pgrep");
+		}
+		/*
+		 * flawfinder: deliberately calling pgrep as there isn't a
+		 * portable library to use instead.
+		 */
+		exit(EXIT_FAILURE);
+	}
+
+	/* Close the write end of the pipe. */
+	(void) close(fds[1]);
+
+	/* Open a file stream on the read end of the pipe. */
+	fptr = fdopen(fds[0], "r");
+	if (NULL == fptr) {
+		perror("fdopen");
+		(void) close(fds[0]);
+		return false;
+	}
+
+	/* Read the lines of output from the child process. */
+
+	ok = true;
+	memset(buffer, 0, sizeof(buffer));
+	while ((0 == feof(fptr)) && (NULL != fgets(buffer, (int) (sizeof(buffer) - 1), fptr))) {
+		unsigned int watch_pid;
+
+		/* Skip all lines if we've hit any errors. */
+		if (!ok) {
+			memset(buffer, 0, sizeof(buffer));
+			continue;
+		}
+
+		/* Skip lines without valid PID. */
+		watch_pid = 0;
+		if (sscanf(buffer, "%u", &watch_pid) < 1)
+			continue;
+		if (watch_pid < 1)
+			continue;
+
+		if (!opts_watchfd_parse(opts, buffer, NULL, 0))
+			ok = false;
+
+		memset(buffer, 0, sizeof(buffer));
+	}
+
+
+	/* Close the stream from the read end of the pipe. */
+	(void) fclose(fptr);
+
+	/* Wait for the child process to exit. */
+	do {
+		/*@-type@ */
+		/* splint disagreement about __pid_t vs pid_t. */
+		pid_status = 0;
+		waited_pid = waitpid(pid, &pid_status, 0);
+		/*@+type@ */
+	} while (-1 == waited_pid && errno == EINTR);
+
+	return ok;
+}
+
+
+/*
+ * Read additional "-d" values from the given file, returning false on
+ * failure.
+ */
+static bool opts_watchfd_listfile(opts_t opts, const char *filename)
+{
+	FILE *fptr;
+	char buffer[1024];		 /* flawfinder: ignore */
+	unsigned int linenumber;
+
+	/*
+	 * flawfinder: buffer is zeroed before each use, and fgets() is
+	 * passed one less than its size so the string functions in the loop
+	 * always find a null byte at the end.
+	 */
+
+	fptr = fopen(filename, "r");	    /* flawfinder: ignore */
+	/*
+	 * flawfinder note: caller directly controls filename, and we're
+	 * opening the file read-only.
+	 */
+	if (NULL == fptr) {
+		fprintf(stderr, "%s: -d @: %s: %s\n", opts->program_name, filename, strerror(errno));
+		return false;
+	}
+
+	linenumber = 0;
+	memset(buffer, 0, sizeof(buffer));
+
+	while ((0 == feof(fptr)) && (NULL != fgets(buffer, (int) (sizeof(buffer) - 1), fptr))) {
+		char *argument;
+		char *nlptr;
+
+		linenumber++;
+
+		/* Remove trailing carriage returns or newlines. */
+		nlptr = strchr(buffer, '\r');
+		if (NULL != nlptr)
+			nlptr[0] = '\0';
+		nlptr = strchr(buffer, '\n');
+		if (NULL != nlptr)
+			nlptr[0] = '\0';
+
+		/* Ignore leading spaces. */
+		argument = buffer;
+		while (argument[0] != '\0' && (argument[0] == ' ' || argument[0] == '\t'))
+			argument++;
+
+		/* Ignore blank lines or comment lines. */
+		if ((argument[0] == '\0') || (argument[0] == '#')) {
+			memset(buffer, 0, sizeof(buffer));
+			continue;
+		}
+
+		/* Reject "@" lines. */
+		if (argument[0] == '@') {
+			(void) fclose(fptr);
+			return false;
+		}
+
+		if (!opts_watchfd_parse(opts, argument, filename, linenumber)) {
+			(void) fclose(fptr);
+			return false;
+		}
+
+		memset(buffer, 0, sizeof(buffer));
+	}
+
+	(void) fclose(fptr);
+	return true;
+}
+
+
+/*
+ * Parse a "-d" option value and make the appropriate additions to the list
+ * of watchfd items, returning false on failure.
+ */
+static bool opts_watchfd_parse(opts_t opts, const char *argument, /*@null@ */ const char *filename, unsigned int line)
+{
+	unsigned int parse_pid;
+	int parse_fd;
+
+	if ('@' == argument[0]) {
+		/* "-d @FILE" syntax - read more values from FILE. */
+
+		/* Don't allow this syntax in a list file. */
+		if (NULL != filename) {
+			/*@-mustfreefresh@ *//* see above */
+			fprintf(stderr, "%s: -d @: %s:%u: %s\n",
+				opts->program_name, filename, line, _("list files may not contain @ lines"));
+			return false;
+			/*@+mustfreefresh@ */
+		}
+
+		return opts_watchfd_listfile(opts, argument + 1);
+
+	} else if ('=' == argument[0]) {
+		/* "-d =NAME" syntax - find processes named NAME. */
+		return opts_watchfd_processname(opts, argument + 1);
+	}
+
+	/* "-d PID[:FD]" syntax. */
+
+	parse_pid = 0;
+	parse_fd = -1;
+
+	if (sscanf(argument, "%u:%d", &parse_pid, &parse_fd) < 1) {
+		/*@-mustfreefresh@ *//* see above */
+		if (NULL != filename) {
+			fprintf(stderr, "%s: -d: %s:%u: %s\n",
+				opts->program_name, filename, line, _("process ID or pid:fd pair expected"));
+		} else {
+			fprintf(stderr, "%s: -d: %s\n", opts->program_name, _("process ID or pid:fd pair expected"));
+		}
+		return false;
+		/*@+mustfreefresh@ */
+	}
+
+	if (parse_pid < 1) {
+		/*@-mustfreefresh@ *//* see above */
+		if (NULL != filename) {
+			fprintf(stderr, "%s: -d: %s:%u: %u: %s\n",
+				opts->program_name, filename, line, parse_pid, _("invalid process ID"));
+		} else {
+			fprintf(stderr, "%s: -d: %u: %s\n", opts->program_name, parse_pid, _("invalid process ID"));
+		}
+		return false;
+		/*@+mustfreefresh@ */
+	}
+
+	return opts_watchfd_add_item(opts, (pid_t) parse_pid, parse_fd);
+}
+
 
 /*
  * Set opts->size from the size of the file whose name is size_file,
@@ -134,12 +443,7 @@ static bool opts_use_size_of_file(opts_t opts, const char *size_file)
 	stat_rc = stat(size_file, &sb);
 
 	if (0 != stat_rc) {
-		/*@-mustfreefresh@ */
-		/*
-		 * splint note: the gettext calls made by _() cause memory
-		 * leak warnings, but in this case it's unavoidable, and
-		 * mitigated by the fact we only translate each string once.
-		 */
+		/*@-mustfreefresh@ *//* see above */
 		fprintf(stderr, "%s: %s: %s: %s\n",
 			opts->program_name, size_file, _("failed to stat file"), strerror(errno));
 		return false;
@@ -308,6 +612,8 @@ opts_t opts_parse(unsigned int argc, char **argv)
 		{ "stop-at-size", 0, NULL, (int) 'S' },
 		{ "sync", 0, NULL, (int) 'Y' },
 		{ "direct-io", 0, NULL, (int) 'K' },
+		{ "sparse", 0, NULL, (int) 'O' },
+		{ "sparse-output", 0, NULL, (int) 'O' },
 		{ "discard", 0, NULL, (int) 'X' },
 		{ "store-and-forward", 1, NULL, (int) 'U' },
 		{ "remote", 1, NULL, (int) 'R' },
@@ -323,14 +629,14 @@ opts_t opts_parse(unsigned int argc, char **argv)
 	/*@+nullassign@ */
 	int option_index = 0;
 #endif				/* HAVE_GETOPT_LONG */
-	char *short_options = "hVpteIrab8kTA:fvnqcWD:s:gl0i:w:H:N:u:F:x:L:B:CEZ:SYKXU:R:P:d:m:o:"
+	char *short_options = "hVpteIrab8kTA:fvnqcWD:s:gl0i:w:H:N:u:F:x:L:B:CEZ:SYKOXU:R:P:d:m:o:"
 #ifdef ENABLE_DEBUGGING
 	    "!:"
 #endif
 	    ;
 	int c, numopts;
-	unsigned int check_pid, parse_pid;
-	int check_fd, parse_fd;
+	unsigned int check_pid;
+	int check_fd;
 	opts_t opts;
 	char *leafptr;
 
@@ -370,10 +676,9 @@ opts_t opts_parse(unsigned int argc, char **argv)
 
 	numopts = 0;
 
+	opts->action = PV_ACTION_TRANSFER;
 	opts->interval = 1;
 	opts->delay_start = 0;
-	opts->watch_pid = 0;
-	opts->watch_fd = -1;
 	opts->average_rate_window = 30;
 
 	opts->width_set_manually = false;
@@ -451,8 +756,49 @@ opts_t opts_parse(unsigned int argc, char **argv)
 			}
 			break;
 		case 'd':
+			if ('@' == optarg[0]) {
+				/* "-d @FILE" syntax - check FILE exists. */
+				if (optarg[1] == '\0') {
+					/*@-mustfreefresh@ *//* see above */
+					fprintf(stderr, "%s: -%c @: %s\n",
+						opts->program_name, c, _("missing filename"));
+					opts_free(opts);
+					return NULL;
+					/*@+mustfreefresh@ */
+				} else if (0 != access(optarg + 1, R_OK)) {	/* flawfinder: ignore */
+					/*
+					 * flawfinder rationale: we're not
+					 * using access() to check
+					 * permissions, only to help give a
+					 * usable error message, so there's
+					 * no TOCTOU issue.
+					 */
+					/*@-mustfreefresh@ *//* see above */
+					fprintf(stderr, "%s: -%c @: %s: %s\n",
+						opts->program_name, c, optarg + 1, strerror(errno));
+					opts_free(opts);
+					return NULL;
+					/*@+mustfreefresh@ */
+				}
+				break;
+			} else if ('=' == optarg[0]) {
+				/*
+				 * "-d =NAME" syntax - check NAME is not
+				 * blank.
+				 */
+				if (optarg[1] == '\0') {
+					/*@-mustfreefresh@ *//* see above */
+					fprintf(stderr, "%s: -%c %c: %s\n",
+						opts->program_name, c, optarg[0], _("missing process name"));
+					opts_free(opts);
+					return NULL;
+					/*@+mustfreefresh@ */
+				}
+				break;
+			}
 			if (sscanf(optarg, "%u:%d", &check_pid, &check_fd)
 			    < 1) {
+				/* "-d PID[:FD]" syntax - check numbers. */
 				/*@-mustfreefresh@ *//* see above */
 				fprintf(stderr, "%s: -%c: %s\n",
 					opts->program_name, c, _("process ID or pid:fd pair expected"));
@@ -478,11 +824,11 @@ opts_t opts_parse(unsigned int argc, char **argv)
 		switch (c) {
 		case 'h':
 			display_help();
-			opts->do_nothing = true;
+			opts->action = PV_ACTION_NOTHING;
 			return opts;	    /* early return */
 		case 'V':
 			display_version();
-			opts->do_nothing = true;
+			opts->action = PV_ACTION_NOTHING;
 			return opts;	    /* early return */
 		case 'p':
 			opts->progress = true;
@@ -628,6 +974,10 @@ opts_t opts_parse(unsigned int argc, char **argv)
 		case 'K':
 			opts->direct_io = true;
 			break;
+		case 'O':
+			opts->sparse_output = true;
+			opts->no_splice = true;
+			break;
 		case 'X':
 			opts->discard_input = true;
 			opts->no_splice = true;
@@ -639,6 +989,7 @@ opts_t opts_parse(unsigned int argc, char **argv)
 				opts_free(opts);
 				return NULL;
 			}
+			opts->action = PV_ACTION_STORE_AND_FORWARD;
 			break;
 		case 'R':
 			opts->remote = (pid_t) pv_getnum_count(optarg, false);
@@ -668,12 +1019,11 @@ opts_t opts_parse(unsigned int argc, char **argv)
 			}
 			break;
 		case 'd':
-			parse_pid = 0;
-			parse_fd = -1;
-			/* No syntax check here, already done earlier */
-			(void) sscanf(optarg, "%u:%d", &parse_pid, &parse_fd);
-			opts->watch_pid = (pid_t) parse_pid;
-			opts->watch_fd = parse_fd;
+			if (!opts_watchfd_parse(opts, optarg, NULL, 0)) {
+				opts_free(opts);
+				return NULL;
+			}
+			opts->action = PV_ACTION_WATCHFD;
 			break;
 		case 'o':
 			opts->output = pv_strdup(optarg);
@@ -722,7 +1072,7 @@ opts_t opts_parse(unsigned int argc, char **argv)
 	if (NULL == opts)
 		return NULL;
 
-	if (0 != opts->watch_pid) {
+	if (PV_ACTION_WATCHFD == opts->action) {
 		if (opts->linemode || opts->null_terminated_lines || opts->stop_at_size
 		    || (opts->skip_errors > 0) || (opts->buffer_size > 0)
 		    || (opts->rate_limit > 0)) {
@@ -752,14 +1102,24 @@ opts_t opts_parse(unsigned int argc, char **argv)
 			/*@+mustfreefresh@ */
 		}
 
-		if (optind < (int) argc) {
+		if (NULL != opts->output) {
 			/*@-mustfreefresh@ *//* see above */
-			fprintf(stderr, "%s: %s\n", opts->program_name,
+			fprintf(stderr, "%s: -o: %s\n", opts->program_name,
 				_("cannot transfer files when watching file descriptors"));
 			opts_free(opts);
 			return NULL;
 			/*@+mustfreefresh@ */
 		}
+
+		/* Accept additional watchfd arguments. */
+		while (optind < (int) argc) {
+			if (!opts_watchfd_parse(opts, argv[optind], NULL, 0)) {
+				opts_free(opts);
+				return NULL;
+			}
+			optind++;
+		}
+
 #ifndef __APPLE__
 		if (0 != access("/proc/self/fdinfo", X_OK)) {	/* flawfinder: ignore */
 			/*
